/**************************Start of CRCMAN.C *************************
 * vi: set sw=4 expandtab:
 *
 * This program is used to build a list of CRC-32 values for all of the
 * files in a given directory tree.  After building the file, the program
 * can be run later to verify the CRC values, giving assurance of the
 * integrity of the files.  To build the CRC file, the command line is:
 *
 *   CRCMAN -b root-dir crc-file-name
 *
 * To check the list of files created, run with this command line:
 *
 *   CRCMAN crc-file-name
 *
 * For a complete list of options see PrintUsage() further down or run the
 * compiled program without parameters.
 *
 * This program should work with most 16 and 32 bit compilers under MS-DOS and
 * UNIX. Use -DDOS to compile under DOS.
 */

/*
 * Original source code by Mark Nelson
 * Dr. Dobb's Journal May, 1992
 * See accompanying file crcman.htm
 *
 * Enhancements by Eike Rathke (erAck) <er@erack.de>
 * Latest version available at http://www.erack.de/download/
 *
 * 01Jun2001: compile with gcc
 * 02Jun2001: removed arbitrary limit of 81 (?!?) characters per filename
 *            in UNIX version (set to PATH_MAX if known or 260 if not)
 * 02Jun2001: check the CRC of one file specified on the command line
 * 02Jun2001: fatal error if write to CRC file failed
 * 03Jun2001: added option -a, append CRCs of new files to crc-file
 * 03Jun2001: bugfix of the original source code in ProcessAllFiles();
 *            needed a closedir() for UNIX, otherwise would run out of
 *            directory handles
 * 04Jun2001: added option -f, use full path names instead of relative paths
 * 04Jun2001: safer handling of crc log file scan
 * 04Jun2001: added option -r, restrict scan to specified directory
 * 04Jun2001: bugfix of the original source code in CheckFiles();
 *            if CRCs didn't match, the printf() had one argument too much,
 *            causing wrong CRC values being displayed
 * 06Jun2001: added @input_dir, a file containing a list of directories
 * 08Jun2001: added option -x exclude_file, a file containing a list of names
 *            (directories and files) to be excluded.
 *            On UNIX, the /dev, /dev/inet and /usr/local/ftp/dev directories
 *            are always excluded.
 * 09Jun2001: better command line checking; check if given file names are
 *            directories and abort if so (since you can open a directory on
 *            UNIX)
 * 09Jun2001: fixed a possible buffer overflow in crc log file scan, and long
 *            file names containing spaces are ok now
 * 29Jun2001: got rid of that ugly K&R style and changed to real prototyping
 * 30Jun2001: POSIX opendir()/readdir()/closedir() also for DJGPP results in
 *            long file names being enabled under win9x, and scan of hidden and 
 *            system files. Also _dos_findfirst()/_dos_findnext() for other
 *            compilers now scan for hidden and system files. Don't know TURBOC
 *            though, so you'll have to change that yourself.
 * 01Jul2001: all error messages now contain the string "Error" (capital E) so
 *            you may grep Error instead of grep -i error and won't get
 *            confused by file names containing the string error anymore
 * 01Jul2001: changed exclude file behavior to work like expected if -f option
 *            was given, and directories without path are excluded as files
 *            are. Also trailing empty lines are ok now.
 * 06Jul2001: added check of entire subdirectory hierarchy if given file name
 *            to check ends in a trailing directory separator
 * 06Jul2001: if the check expects a file but encounters a directory under the
 *            same name instead, it's a normal Error now instead of a Fatal
 *            Error
 * 07Jul2001: added option -c, clean CRC file of nonexisting files
 *            added option -e together with -c, clean also any files with
 *            errors in the CRC check sum
 * 07Jul2001: added option -d, delete a specified entry or subdirectory
 *            hierarchy
 * 08Jul2001: multiple files and directories may be specified for
 *            check/clean/delete. On UNIX, shell expands wildcards, under DOS
 *            DJGPP does it internally. Don't know for TURBOC and other DOS
 *            compilers, except that with MSC you must link a special object
 *            file to achieve this, forgot its name.
 * 08Jul2001: changed -b and -a options to take the same order of crc-file and
 *            input directory names as all other options. Multiple input
 *            directories may be specified.
 * 15Jul2001: speeded up directory check/clean/delete by using start and stop
 *            positions in the sorted CRC array
 * 21Jul2001: under DOS, all file name comparisons are case insensitive now,
 *            no need for -f option for this case anymore, was confusing
 *            anyways
 * 22Jul2001: added option -p, pervert path under DOS ('/' instead of '\' and
 *            lowercase file names and no carriage returns are written to CRC
 *            file)
 * 22Jul2001: added @ to file-to-check and files-to-delete for files containing
 *            a list of names
 * 22Jul2001: if @listfile can't be opened try to continue with "@listfile" as
 *            the real name
 * 28Jul2001: added option -w, working directory prepended to names
 * 27Oct2002: added option -q quiet, no progress output, and -qq very quiet, no
 *            progress or info output at all, only errors.
 *            Changed absence of -DUNIX to default to -DUNIX instead,
 *            use -DDOS to compile under DOS. This is much more convenient..
 * 05Jan2003: build -b and append -a now also take single file names as
 *            arguments, not just directories, so you may pass a list of names
 *            of files to scan.
 * 03Sep2003: scan and check dangling links (no target file or directory) on
 *            Unix by reading the link value. Don't scan sockets or pipes.
 * 05Sep2003: added option -i, ignore file errors during scan and continue.
 * 06Apr2005: made stdarg.h the default also on Unix systems, define
 *            USE_VARARGS if your compiler doesn't understand it.
 *
 *
 * TODO:
 *
 * Use hashtables instead of sorted arrays, would speed things up significantly
 * on large CRC files, but would also bloat the code. And wouldn't output a
 * nicely sorted CRC file on options -c and -d
 *
 * Implement MD5
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>


/*
 * To build this program under UNIX, define UNIX either on the command
 * line or by editing this file.  To define it on the command line, the
 * program should be built like this:
 *
 *  cc -o crcman -DUNIX crcman.c
 *
 * The code in this program assumes that the UNIX compiler is of the
 * K&R variety, and does away with real function prototyping.
 *
 * erAck: changed this antique dinosaur behavior to real prototyping. If your
 * compiler doesn't support this you're out of luck anyway.
 */

/* erAck: make compile under UNIX default, DOS'ers have to -DDOS */
#ifndef DOS
#define UNIX
#endif

/* erAck: define USE_VARARGS in case your compiler doesn't support stdarg.h but
 * uses varargs.h instead */
#ifdef USE_VARARGS
#include <varargs.h>
#else
#include <stdarg.h>
#endif
 
#ifdef UNIX

/* erAck: opendir(), readdir(), closedir() not restricted to UNIX */
#define USE_POSIX_DIR 1

/* erAck: having lstat() is not very common */
#if defined(__GNUC__)
/* On systems supporting lstat(), dangling links (pointing to a non-existent
 * entry) are not skipped or considered to be an error. Instead readlink() is
 * used, and the link's content is scanned/checked. */
#define USE_LSTAT 1
#else
#define USE_LSTAT 0
#endif
 
#ifdef M_XENIX
#include <sys/ndir.h>
#elif defined(__GNUC__)
/* erAck */
#include <dirent.h>
#else
#include <sys/dirent.h>
#endif /* M_XENIX */

/* erAck: realpath() */
#include <unistd.h>

#define SEPARATOR "/"
#define PERVERT_SEPARATOR SEPARATOR
#define PERVERTIBLE 0
#define PERVERT_SWITCH_USAGE ""
#define PERVERT_SWITCH_ALLOWED 0

#define FILENAMECMP strcmp
#define FILENAMENCMP strncmp

/* erAck: 81 ?!?
#define FILENAME_SIZE 81
*/
#if defined(PATH_MAX)
#define FILENAME_SIZE PATH_MAX
#else
#define FILENAME_SIZE 260
#endif

#else /* not UNIX, must be MSDOS */
/*
 * Most MS-DOS compilers have converged on the same names for the
 * structures and functions used when searching directories.
 * Unfortunately, Borland C implementations still use a variant,
 * which requires a few macro definitions to work around.  The
 * functions work in an identical manner, so the actual
 * implementation of the code is straightforward.  The addition of
 * the MSDOS definition helps convince the Zortech compiler to use
 * the same structure and function names as everyone else.
 *
 */

#define MSDOS 1

#define SEPARATOR "\\"

#define FILENAMECMP stricmp
#define FILENAMENCMP strnicmp

#if defined(__DJGPP)

/* erAck: DJGPP is a well behaving DOS compiler and knows about POSIX,
 * use long file names under win9x */
#define USE_POSIX_DIR 1
#include <dirent.h>
/* __OPENDIR_FIND_HIDDEN is only for compatibilty with DJGPP versions <2.02 */
int __opendir_flags = __OPENDIR_PRESERVE_CASE | __OPENDIR_FIND_HIDDEN;
/* access() */
#include <unistd.h>

#else   /* __DJGPP */

#define USE_POSIX_DIR 0
#include <dos.h>

#endif  /* __DJGPP */

/* DOS compiler libraries understanding '/' as path separator */
#if defined(__DJGPP) || defined(_MSC_VER)
#define PERVERTIBLE 1
#else
#define PERVERTIBLE 0
#endif
#define PERVERT_SEPARATOR "/"
#define PERVERT_SWITCH_USAGE "[-p] "
#define PERVERT_SWITCH_ALLOWED 1

#define FILENAME_SIZE FILENAME_MAX

#ifdef __TURBOC__

#include <dir.h>
#define FILE_INFO                 struct ffblk
/* erAck: don't know TC flags for hidden and system files, do it yourself */
#define FIND_FIRST( n, i )        findfirst( ( n ), ( i ), FA_DIREC )
#define FIND_NEXT( info )         findnext( ( info ) )
#define FILE_NAME( info )         ( ( info ).ff_name )

#else

#define FILE_INFO                 struct find_t
/* erAck: why not scan for hidden and system files?
#define FIND_FIRST( n, i )        _dos_findfirst( (n), _A_SUBDIR, (i) )
*/
#define FIND_FIRST( n, i )        _dos_findfirst( (n), _A_SUBDIR | _A_HIDDEN | _A_SYSTEM, (i) )
#define FIND_NEXT( info )         _dos_findnext( ( info ) )
#define FILE_NAME( info )         ( ( info ).name )

#endif

#endif  /* UNIX */


unsigned long CRCTable[ 256 ];


/* erAck:
 * struct, sort function and two variables for the CRC file sorted by
 * file names.
 */
struct CrcLine {
    unsigned long   crc;
    char            *name;
    int             flag;
};
typedef struct CrcLine CrcLine;
int CrcLineSortFunction( const void* p1, const void* p2 )
{
    return( FILENAMECMP( ((const CrcLine*)p1)->name,
                ((const CrcLine*)p2)->name ) );
}
CrcLine *pSortedCrcArray = NULL;
long nSortedCrcEntries = 0;


/* erAck:
 * sort function and two variables for the exclude list sorted by names.
 */
int ExcludeSortFunction( const void* p1, const void* p2 )
{
    return( FILENAMECMP( *(const char**)p1, *(const char**)p2 ) );
}
char **pSortedExcludeArray = NULL;
long nSortedExcludeEntries = 0;


/* erAck:
 * A struct holding all command line options, all initialized to 0.
 */
struct {
    int build;      /* -b or -a build CRC file */
    int append;     /* -a append CRC of new files to crc log file */
    int restrict;   /* -r restrict to specified directory, no recursion */
    int fullpath;   /* -f use full path name instead of relative path name */
    int clean;      /* -c clean CRC file of nonexisting files */
    int cleanerr;   /* -e clean also entries which produce errors */
    int delete;     /* -d delete entry */
    int pervert;    /* -p pervert paths (`/' instead of `\' and lowercase) */
    int quiet;      /* -q quiet and -qq, no progress, only errors */
    int ignore;     /* -i ignore file access errors and continue */
    char *exclude;  /* -x exclude list file name */
    char *crcfile;  /* name of CRC file */
    char **files;   /* start of argv with names to build/append/check/clean/delete */
    int  filecount; /* how many file names to build/append/check/clean/delete */
    char cwd[FILENAME_SIZE];    /* current working directory */
} options;

#ifdef USE_VARARGS
/* variadic functions are the only old-style left-over */
void FatalError();
void OutProgress();
void OutInfo();
#else
void FatalError( char *fmt, ... );
void OutProgress( char *fmt, ... );
void OutInfo( char *fmt, ... );
#endif
unsigned long CalculateFileCRC( FILE *file );
void ProcessAllFiles( char *path, FILE *crc_file, int bIsInRecursion );
int ProcessOneFile( FILE* crc_file, char* name );
void BuildCRCFile( char *crc_file_name, char **dir_names, int dir_count );
void WriteCrcLog( FILE *crc_file, unsigned long crc, char *name );
void CreateCRCFileArray( FILE *crc_file );
const CrcLine * SearchCRCFileArray( char *file_name, long *pInsert );
void DeleteCRCFileArray();
void CreateExcludeArray();
const char * SearchExcludeArray( const char *name );
void DeleteExcludeArray();
void CheckFiles( char *crc_file_name, char **file_names, int file_count );
unsigned long CalculateBufferCRC( unsigned int count, unsigned long crc,
                                  void *buffer );
void BuildCRCTable( void );
char * GetFullPathName( char *fullpath, char *name );
char * PervertPath( char *pPath );
int IsDir( const char *name, int file_needed, int must_stat, mode_t *nStatMode );
int Exists( const char* name );
char * PrependCwdPath( char* pName );


/* erAck:
 * Print usage.
 */

void PrintUsage( char* prog )
{
    printf( "Usage: %s -b [-r] %s[-x exclude_file] [-w] [-f] [-q] [-i] crc-file [@]input_dirs... \n", prog, PERVERT_SWITCH_USAGE );
    printf( "       %s -a [-r] %s[-x exclude_file] [-w] [-f] [-q] [-i] crc-file [@]input_dirs... \n", prog, PERVERT_SWITCH_USAGE );
    printf( "\n" );
    printf( "-b (build) scans all files under input_dirs\n" );
    printf( "   and writes their data to the crc-file. crc-file is overwritten.\n" );
    printf( "-a (append) scans all files under input_dirs\n" );
    printf( "   and appends their data to the crc-file if not already present.\n" );
    printf( "\n" );
    printf( "input_dirs may be any number (limited by your command line capacity of course)\n" );
    printf( "of directories or files to scan.\n" );
    printf( "\n" );
    printf( "-r (restrict) with -b or -a restricts the scan\n" );
    printf( "   to the specified directories, no recursion is done.\n" );
    printf( "\n" );
    printf( "   If @input_dir is used, input_dir is taken as the file name of a\n" );
    printf( "   file containing a list of directories or files. If a directory is followed\n" );
    printf( "   by \"  -r\" (exactly two spaces and -r) the -r option is applied\n" );
    printf( "   to that directory. A \"  -r-\" will revoke a -r option supplied\n" );
    printf( "   on the command line for that directory.\n" );
    printf( "\n" );
    printf( "-x (exclude) reads a list of names from exclude_file\n" );
    printf( "   preventing any matching directory or file to be scanned.\n" );
    printf( "   On UNIX, /dev, /dev/inet and /usr/local/ftp/dev are always\n" );
    printf( "   added internally to the list. Guess why ...\n" );
    printf( "\n" );
    printf( "-i ignore file access errors and continue, instead of aborting.\n" );
    printf( "\n" );
    printf( "Usage: %s [-w] [-f] [-q] crc-file [[@]files-to-check...]\n", prog );
    printf( "   If no file-to-check is specified, check the CRC data of all of the\n" );
    printf( "   files in the crc-file and print the results.\n" );
    printf( "   Otherwise, check the CRC data of one or more files against crc-file.\n" );
    printf( "   Or check the CRC data of an entire subdirectory hierarchy if the\n" );
    printf( "   given name ends in a directory separator `%s'\n", SEPARATOR );
    printf( "\n" );
    printf( "Usage: %s -c [-e] %s[-w] [-f] [-q] crc-file [[@]files-to-check...]\n", prog, PERVERT_SWITCH_USAGE );
    printf( "-c Clean the CRC file of nonexisting files.\n" );
    printf( "   Limit to files-to-check if specified, or subdirectory hierarchy if the\n" );
    printf( "   given name ends in a directory separator `%s'\n", SEPARATOR );
    printf( "-e option also cleans entries with errors in CRC sums.\n" );
    printf( "\n" );
    printf( "Usage: %s -d %s[-w] [-f] [-q] crc-file [@]files-to-delete...\n", prog, PERVERT_SWITCH_USAGE );
    printf( "-d Delete one or more entries in the CRC file,\n" );
    printf( "   or an entire subdirectory hierarchy if the\n" );
    printf( "   given name ends in a directory separator `%s'\n", SEPARATOR );
    printf( "\n" );
    printf( "Global switches:\n" );
    printf( "\n" );
    printf( " @ If @filename is used in any case, filename is taken as a file\n" );
    printf( "   containing a list of names, one per line.\n" );
    printf( "\n" );
#if PERVERT_SWITCH_ALLOWED
    printf( "-p (pervert path) in any case will store the file names\n" );
    printf( "   in the CRC log file using `/' as the path separator and converting\n" );
    printf( "   all ASCII letters [A-Z] to lowercase. Additionally, no carriage return\n" );
    printf( "   is written, just a line feed. If you don't use any drive letters in\n" );
    printf( "   input_dir specifications the CRC log file then will also be usable on UNIX.\n" );
#endif
    printf( "-w (working directory) prepends any specified\n" );
    printf( "   file or directory name with the name of the current working directoy.\n" );
    printf( "\n" );
    printf( "-f (fullpath) in any case will store, retrieve\n" );
    printf( "   and compare all file names using their full path names instead of\n" );
    printf( "   path names relative to the specified path. However, using this option\n" );
    printf( "   will slow down the process significantly on large CRC files.\n" );
    printf( "\n" );
    printf( "-q (quiet) suppresses all progress output,\n" );
    printf( "-qq (quiet quiet) all progress and info output is suppressed,\n" );
    printf( "    only errors are reported.\n" );
    printf( "\n" );
    printf( "Again, the usage summary:\n" );
    printf( " Build: %s -b [-r] %s[-x exclude_file] [-w] [-f] [-q] [-i] crc-file [@]input_dirs... \n", prog, PERVERT_SWITCH_USAGE );
    printf( "Append: %s -a [-r] %s[-x exclude_file] [-w] [-f] [-q] [-i] crc-file [@]input_dirs... \n", prog, PERVERT_SWITCH_USAGE );
    printf( " Check: %s [-w] [-f] [-q] crc-file [[@]files-to-check...]\n", prog );
    printf( " Clean: %s -c [-e] %s[-w] [-f] [-q] crc-file [[@]files-to-check...]\n", prog, PERVERT_SWITCH_USAGE );
    printf( "Delete: %s -d %s[-w] [-f] [-q] crc-file [@]files-to-delete...\n", prog, PERVERT_SWITCH_USAGE );
}

/*
 *
 * The main program is fairly simple.  It checks for valid occurences
 * of the different types of command lines, and executes them if
 * found.  Otherwise, it prints out a simple usage statement and exits.
 */

int main( int argc, char* argv[] )
{
    int ret = 0;
    int arg;
    for ( arg=1; arg < argc && !ret; ++arg )
    {
        if ( argv[arg][0] == '-' )
        {
            /* don't allow concatenated switches, except -qq */
            if ( argv[arg][1] && argv[arg][2] && argv[arg][1] != 'q' )
                ret = 1;
            else
            {
                switch ( argv[arg][1] )
                {
                    case 'b':
                        options.build = 1;
                    break;
                    case 'a':
                        options.build = 1;
                        options.append = 1;
                    break;
                    case 'r':
                        options.restrict = 1;
                    break;
                    case 'x':
                        if ( !options.exclude && arg < argc-1 )
                            options.exclude = argv[ ++arg ];
                        else
                            ret = 1;
                    break;
                    case 'i':
                        options.ignore = 1;
                    break;
                    case 'f':
                        options.fullpath = 1;
                    break;
                    case 'c':
                        options.clean = 1;
                    break;
                    case 'e':
                        options.cleanerr = 1;
                    break;
                    case 'd':
                        options.delete = 1;
                    break;
#if PERVERT_SWITCH_ALLOWED
                    case 'p':
                        options.pervert = 1;
                    break;
#endif
                    case 'w':
                    {
                        int nLen;
                        GetFullPathName( options.cwd, "." );
                        nLen = strlen( options.cwd );
                        if ( nLen && options.cwd[ nLen - 1 ] == '.' )
                            options.cwd[ --nLen ] = '\0';
                        if ( nLen && options.cwd[ nLen - 1 ] == SEPARATOR[0] )
                            options.cwd[ --nLen ] = '\0';
                    }
                    break;
                    case 'q':
                        ++options.quiet;
                        if ( argv[arg][2] == 'q' )
                            ++options.quiet;
                    break;
                    default:
                        ret = 1;
                }
            }
        }
        else if ( !options.crcfile )
            options.crcfile = PervertPath( argv[ arg ] );
        else if ( !options.files )
        {
            PervertPath( argv[ arg ] );
            options.files = &argv[ arg ];
            options.filecount = 1;
        }
        else
        {
            PervertPath( argv[ arg ] );
            ++options.filecount;
        }
    }
    if ( !options.crcfile
            || ((options.clean || options.delete)
                && (options.build || options.exclude || options.restrict))
            || (options.cleanerr && !options.clean)
            || (options.delete && (options.clean || !options.files))
            || (options.build && !options.files)
            )
        ret = 1;
    if ( ret )
    {
        PrintUsage( argv[0] );
        return( ret );
    }
    setbuf( stdout, NULL );
    BuildCRCTable();
    if ( options.build )
        BuildCRCFile( options.crcfile, options.files, options.filecount );
    else
        CheckFiles( options.crcfile, options.files, options.filecount );
    return( ret );
}

/*
 * Instead of performing a straightforward calculation of the 32 bit
 * CRC using a series of logical operations, this program uses the
 * faster table lookup method.  This routine is called once when the
 * program starts up to build the table which will be used later
 * when calculating the CRC values.
 */

#define CRC32_POLYNOMIAL     0xEDB88320L

void BuildCRCTable()
{
    int i;
    int j;
    unsigned long crc;

    for ( i = 0; i <= 255 ; i++ ) {
        crc = i;
        for ( j = 8 ; j > 0; j-- ) {
            if ( crc & 1 )
                crc = ( crc >> 1 ) ^ CRC32_POLYNOMIAL;
            else
                crc >>= 1;
        }
        CRCTable[ i ] = crc;
    }
}

/* erAck:
 * Get full path name of a specified (relative) file name.
 * pFullPath must be a buffer capable of holding at least FILENAME_SIZE chars.
 * If for any reason the full path name can't be resolved the relative file
 * name is copied to pFullPath.
 */

char * GetFullPathName( char* pFullPath, char* pName )
{
#if defined(UNIX)
    if ( !realpath( pName, pFullPath ) )
        strcpy( pFullPath, pName );
#elif defined(__DJGPP)
    _fixpath( pName, pFullPath );
#else   /* MS-DOS */
    if ( !_fullpath( pFullPath, pName, FILENAME_SIZE ) )
        strcpy( pFullPath, pName );
#endif
#if PERVERTIBLE
    /* when using _fixpath() DJGPP sets the directory separator to '/' */
    PervertPath( pFullPath );
#endif
    return( pFullPath );
}


/*
 * erAck:
 * Prepend a name with the current working directory if option -w given.
 */
char * PrependCwdPath( char *pName )
{
    if ( *options.cwd )
    {
        char tmp[FILENAME_SIZE];
        char *p = pName;
        if ( p[0] == '.' && (!p[1] || p[1] == SEPARATOR[0]) )
            ++p;    /* "." or "./" */
        strcpy( tmp, p );
        if ( *p && *p != SEPARATOR[0] )
        {
            strcpy( pName, options.cwd );
            strcat( pName, SEPARATOR );
        }
        else
            strcpy( pName, options.cwd );
        strcat( pName, tmp );
    }
    return pName;
}


/* erAck:
 * Pervert a path if not UNIX by replacing every '/' with '\'.
 */

char * PervertPath( char* pPath )
{
#ifndef UNIX
    char *p = pPath;
    while ( *p )
    {
        if ( *p == '/' )
            *p = SEPARATOR[0];
        ++p;
    }
#endif
    return( pPath );
}


/* erAck:
 * Write a CRC logfile entry. Under DOS, a perverted path is written using '\'
 * instead of '/' if option -p was not given, otherwise '/' is written and case
 * is changed to lower case. The buffer passed is changed!
 */

void WriteCrcLog( FILE *crc_file, unsigned long crc, char *name )
{
#if PERVERT_SWITCH_ALLOWED
    if ( options.pervert )
    {
        char* p = name;
        while ( *p )
        {
            if ( 'A' <= *p && *p <= 'Z' )
                *p += 'a' - 'A';
            else if ( *p == SEPARATOR[0] )
                *p = '/';
            ++p;
        }
    }
#if PERVERTIBLE
    else
    {
        char* p = name;
        while ( *p )
        {
            if ( *p == '/' )
                *p = SEPARATOR[0];
            ++p;
        }
    }
#endif
#endif
    if ( fprintf( crc_file, "%08lx %s\n", crc, name ) < 0 )
        FatalError( "Could not write to CRC file!\n" );
}


/* erAck:
 * Get crc value and file name from crc log file. Return number of assigned
 * fields (2) if found, 0 else. If not found all characters up to the next line
 * feed are read, so the calling funtion may continue crc log file processing.
 */

int GetScanLine( FILE* crc_file, unsigned long* log_crc, char* log_name )
{
    int result;
    static char buf[32];
    static int init = 1;
    if ( init )
    {
        /* erAck: bugfix of original source
         * couldn't cope with spaces in file names 
         * and had a possible buffer overflow
         * fscanf( ..., "%lx %s", ... ) */
        sprintf( buf, "%%lx %%%d[^\n]", FILENAME_SIZE-1 );
        init = 0;
    }
    result = fscanf( crc_file, buf, log_crc, log_name );
    if ( result < 2 )
    {
        int len;
        int c = getc( crc_file );
        if ( c == EOF )
            return( 0 );
        log_name[0] = c;
        fgets( &log_name[1], FILENAME_SIZE-1, crc_file );
        len = strlen( log_name ) - 1;
        c = log_name[ len ];
        log_name[ len ] = '\n';
        printf( "Error in crc log file: %s", log_name );
        while ( c != '\n' && c != EOF )
            c = getc( crc_file );
        return( 0 );
    }
    return( result );
}


void CheckOneArgument( FILE* crc_file, char* crc_file_name, char* file_name )
{
    char full_file_name[ FILENAME_SIZE ];
    int found = 0;
    char *pFileName = NULL;
    int bCheckDir = 0;
    int nLen= 0;
    long nSortedEntry = 0;
    long nStopEntry = nSortedCrcEntries;
    int bSorted = (pSortedCrcArray != NULL);

    if ( file_name )
    {
        PervertPath( file_name );
        PrependCwdPath( file_name );
        /* force a directory check if a trailing separator is specified */
        nLen = strlen( file_name );
        bCheckDir = (nLen >= strlen( SEPARATOR ) ?
            strcmp( file_name + nLen - strlen( SEPARATOR ), SEPARATOR ) == 0 :
            0);
        if ( options.fullpath )
        {
            pFileName = GetFullPathName( full_file_name, file_name );
            nLen = strlen( pFileName );
            if ( bCheckDir &&
                    strcmp( pFileName + nLen - strlen( SEPARATOR ),
                        SEPARATOR ) != 0 )
            {   /* append a separator for strncmp() if needed */
                strcat( pFileName, SEPARATOR );
                nLen += strlen( SEPARATOR );
            }
        }
        else
        {
            pFileName = file_name;
            if ( bSorted )
            {
                /* get start position in sorted array; however, no fullpath is
                 * tried since the sorted array contains the real CRC file data
                 */
                const CrcLine* pL = SearchCRCFileArray( pFileName, &nSortedEntry );
                /* get stop position */
                if ( pL && !bCheckDir )
                    nStopEntry = nSortedEntry + 1;
                else if ( nSortedEntry >= nSortedCrcEntries )
                    nStopEntry = nSortedCrcEntries;
                else
                {
                    char ch = pFileName[nLen-1];
                    pFileName[nLen-1] = 0xff;
                    pL = SearchCRCFileArray( pFileName, &nStopEntry );
                    if ( pL )
                        ++nStopEntry;
                    pFileName[nLen-1] = ch;
                }
            }
        }
    }
    while ( bSorted ? nSortedEntry < nStopEntry : !feof( crc_file ) )
    {
        unsigned long log_crc;
        unsigned long crc;
        char log_name[ FILENAME_SIZE ];
        char full_log_name[ FILENAME_SIZE ];
        char *pLogName;

        if ( bSorted )
        {
            log_crc = pSortedCrcArray[ nSortedEntry ].crc;
            strcpy( log_name, pSortedCrcArray[ nSortedEntry ].name );
        }
        else
        {
            if ( !GetScanLine( crc_file, &log_crc, log_name ) )
                continue;
            PervertPath( log_name );
        }
        pLogName = (options.fullpath ?
                GetFullPathName( full_log_name, log_name ) :
                log_name);
        if ( !pFileName || (bCheckDir ? 
                    FILENAMENCMP( pFileName, pLogName, nLen ) == 0 :
                    FILENAMECMP( pFileName, pLogName ) == 0) )
        {
            found = 1;
            if ( options.delete )
            {
                OutInfo( "Removing entry %s\n", pLogName );
                pSortedCrcArray[ nSortedEntry ].flag = 1;
            }
            else if ( IsDir( pLogName, 0, 0, NULL ) )
            {   /* IsDir() must not stat if file doesn't exist */
                printf( "Error: Needed file, not directory: %s\n", pLogName );
                if ( options.cleanerr )
                {
                    OutInfo( "Removing mismatching entry %s\n", pLogName );
                    pSortedCrcArray[ nSortedEntry ].flag = 1;
                }
            }
            else
            {
                if ( options.clean && !Exists( pLogName ) )
                {
                    OutInfo( "Removing nonexistent entry %s\n", pLogName );
                    pSortedCrcArray[ nSortedEntry ].flag = 1;
                }
                else if ( (!options.clean && !options.delete)
                        || options.cleanerr )
                {
                    FILE* test_file = fopen( pLogName, "rb" );
                    if ( test_file != NULL )
                    {
                        OutProgress( "Checking %s ", pLogName  );
                        crc = CalculateFileCRC( test_file );
                        fclose( test_file );
                        if ( crc != log_crc )
                        {
                            if ( options.quiet < 1 )
                                printf( "Error:  Expected %08lx, got %08lx\n",
                                        log_crc, crc );
                            else
                                printf( "Error: %s: Expected %08lx, got %08lx\n",
                                        pLogName, log_crc, crc );
                            if ( options.cleanerr )
                            {
                                OutInfo( "Removing mismatching entry %s\n",
                                        pLogName );
                                pSortedCrcArray[ nSortedEntry ].flag = 1;
                            }
                        }
                        else
                            OutProgress( "OK\n" );
                    }
                    else
                    {
                        int ret = 0;
                        struct stat buf;
                        int fileerrno = errno;
#if USE_LSTAT
                        int lbad = (lstat( pLogName, &buf ) == -1);
                        if ( !lbad && S_ISLNK(buf.st_mode) )
                        {
                            char target[ FILENAME_SIZE ];
                            int len;
                            OutInfo( "Checking link only: %s ", pLogName );
                            len = readlink( pLogName, target, FILENAME_SIZE-1 );
                            if ( len < 0 )
                                OutInfo( "(can't resolve, %s)\n",
                                        strerror( errno ) );
                            else
                            {
                                unsigned long crc = 0xFFFFFFFFL;
                                target[ len ] = '\0';
                                OutInfo( "-> %s ", target );
                                crc = CalculateBufferCRC( len, crc, target );
                                /* NOT xor'ing out with 0xFFFFFFFFL to
                                 * distinguish from identical file content */
                                if ( crc != log_crc )
                                {
                                    if ( options.quiet < 2 )
                                        printf( "Error:  Expected %08lx, got %08lx\n",
                                                log_crc, crc );
                                    else
                                        printf( "Error: %s: Expected %08lx, got %08lx\n",
                                                pLogName, log_crc, crc );
                                    if ( options.cleanerr )
                                        OutInfo( "NOT removing mismatching link entry %s\n",
                                                pLogName );
                                }
                                else
                                    OutInfo( "OK\n" );
                                ret = 1;
                            }
                        }
#endif // USE_LSTAT
                        if ( !ret )
                            printf( "Error: Could not open %s: %s\n", pLogName,
                                    strerror( fileerrno ));
                    }
                }
            }
            if ( pFileName && !bCheckDir )
                break;  /* while */
        }
        ++nSortedEntry;
    }
    if ( pFileName && !found )
        printf( "Error: %s not found in crc log file: %s\n", 
                pFileName, crc_file_name );
}


/*
 * The routine to check the CRC values for a list of files has a
 * fairly easy job of it.  It just reads in a line at a time from
 * the CRC file.  Each line contains a file name and a CRC value.
 * The program then just has to calculate the actual CRC for that
 * file, and compare it with the current calculated value.  Any
 * discrepancy triggers an error message.
 *
 * erAck: it's not that fairly easy anymore since we can check a file name, an
 * entire subdirectory hierarchy, or a list thereof.
 */

void CheckFiles( char* crc_file_name, char** file_names, int file_count )
{
    FILE *crc_file;
    int bRewrite = (options.clean || options.delete);
    int bSorted = (bRewrite || file_count > 1 ||
            (file_count == 1 && file_names[0][0] == '@'));

    IsDir( crc_file_name, 1, 1, NULL );
    crc_file = fopen( crc_file_name, "r" );
    if ( crc_file == NULL )
        FatalError( "Couldn't open the CRC log file for reading: %s\n",
                crc_file_name );
    if ( bSorted )
        CreateCRCFileArray( crc_file );
    if ( !file_names )
       CheckOneArgument( crc_file, crc_file_name, NULL );
    else
    {
        int nFile;
        for ( nFile=0; nFile < file_count; ++nFile )
        {
            FILE *list_file;
            char *pName = file_names[nFile];
            int nLen = strlen( pName );
            /* may be a @file containing a list of names */
            if ( nLen > 1 && pName[0] == '@' )
            {
                /* must not stat if doesn't exist */
                IsDir( &pName[1], 1, 0, NULL );
                list_file = fopen( &pName[1], "r" );
                if ( !list_file )
                    printf( "Error: Can't open list file: %s\n", &pName[1] );
                    /* continue and try to take it as a real file */
            }
            else
                list_file = NULL;
            if ( !list_file )
                CheckOneArgument( crc_file, crc_file_name, pName );
            else
            {
                do 
                {
                    char file_name[ FILENAME_SIZE ];
                    int nLen;
                    if ( !fgets( file_name, FILENAME_SIZE, list_file ) )
                    {
                        if ( feof( list_file ) )
                            break;
                        FatalError( "Could not read from list file: %s\n",
                                list_file );
                    }
                    nLen = strlen( file_name );
                    if ( nLen && file_name[nLen-1] == '\n' )
                        file_name[--nLen] = '\0';
                    CheckOneArgument( crc_file, crc_file_name, file_name );
                } while ( !feof( list_file ) );
                fclose( list_file );
            }
        }
    }
    fclose( crc_file );
    if ( bRewrite )
    {
        long j;
        /* a safety check; if every entry was flagged to be deleted there was
         * propably a relative path in the CRC file that didn't match the
         * current directory */
        for ( j=0; j < nSortedCrcEntries; ++j )
        {
            if ( !pSortedCrcArray[j].flag )
                break;
        }
        if ( j >= nSortedCrcEntries )
            printf( "Nothing matched, for safety not removing anything.\n" );
        else
        {
            crc_file = fopen( crc_file_name, (options.pervert ? "wb" : "w") );
            if ( crc_file == NULL )
                FatalError( "Couldn't open the CRC log file for writing: %s\n",
                        crc_file_name );
            for ( j=0; j < nSortedCrcEntries; ++j )
            {
                if ( !pSortedCrcArray[j].flag )
                    WriteCrcLog( crc_file, pSortedCrcArray[j].crc,
                            pSortedCrcArray[j].name );
            }
            fclose( crc_file );
        }
    }
    if ( bSorted )
        DeleteCRCFileArray();
}


/* erAck:
 * Create a sorted array of the CRC file entries
 */
void CreateCRCFileArray( FILE* crc_file )
{
    unsigned long log_crc;
    char log_name[ FILENAME_SIZE ];
    fseek( crc_file, 0, SEEK_SET );
    while ( !feof( crc_file ) && !ferror( crc_file ) ) 
    {
        if ( !GetScanLine( crc_file, &log_crc, log_name ) )
            continue;
        /* this is more hypothetical ... */
        if ( nSortedCrcEntries == LONG_MAX )
            FatalError( "Too many entries in CRC file: %ld\n",
                    nSortedCrcEntries );
        ++nSortedCrcEntries;
    }
    if ( nSortedCrcEntries )
    {
        int result;
        long j;
        pSortedCrcArray = 
            (CrcLine*) malloc( nSortedCrcEntries * sizeof(CrcLine) );
        if ( !pSortedCrcArray )
            FatalError( "Can't allocate sorted array of CRC file.\n" );
        fseek( crc_file, 0, SEEK_SET );
        for ( j=0; j < nSortedCrcEntries && !feof( crc_file )
                && !ferror( crc_file ); ++j ) 
        {
            if ( !GetScanLine( crc_file, &log_crc, log_name ) )
            {
                --j;
                continue;
            }
            PervertPath( log_name );
            pSortedCrcArray[j].crc = log_crc;
            pSortedCrcArray[j].name = strdup( log_name );
            pSortedCrcArray[j].flag = 0;
            if ( !pSortedCrcArray[j].name )
                FatalError( "Can't allocate entry in sorted CRC array.\n" );
        }
        result = GetScanLine( crc_file, &log_crc, log_name );
        if ( j < nSortedCrcEntries || result || !feof( crc_file )
                || ferror( crc_file ) )
            FatalError( "Error in reading crc log file.\n" );
        qsort( pSortedCrcArray, nSortedCrcEntries, sizeof(CrcLine),
            &CrcLineSortFunction );
    }
    if ( !nSortedCrcEntries )
    {
        fclose( crc_file );
        FatalError( "Could not read any entries from crc log file.\n" );
    }
    fseek( crc_file, 0, SEEK_END );
}


/* erAck:
 * Search a sorted array of CRC file entries for a filename
 */
const CrcLine * SearchCRCFileArray( char* file_name, long* pInsert )
{
    long hi, lo, mid;
    int cmp = 0;
    hi = nSortedCrcEntries - 1;
    lo = mid = 0;
    while ( lo <= hi )
    {
        /* safer against overflows than ((lo+hi)/2) */
        mid = lo + ((hi - lo) / 2);
        cmp = FILENAMECMP( pSortedCrcArray[mid].name, file_name );
        if ( cmp < 0 )
            lo = mid + 1;
        else if ( cmp > 0 )
            hi = mid - 1;
        else
        {
            if ( pInsert )
                *pInsert = mid;
            return( &pSortedCrcArray[mid] );
        }
    }
    if ( pInsert )
        *pInsert = (cmp < 0 ? mid+1 : mid);
    return( NULL );
}


void DeleteCRCFileArray()
{
    if ( pSortedCrcArray )
    {
        long j;
        for ( j=0; j < nSortedCrcEntries; ++j )
        {
            free( pSortedCrcArray[j].name );
        }
        free( pSortedCrcArray );
        pSortedCrcArray = NULL;
        nSortedCrcEntries = 0;
    }
}


/* erAck:
 * Create a sorted array of exclude entries
 */
void CreateExcludeArray()
{
    static char *pAlways[] = {
#ifdef UNIX
        /* never check /dev/zero, /dev/full, /dev/ram*, /dev/random ... */
        "/dev",
        "/dev/inet",
        "/usr/local/ftp/dev",
#endif
        NULL
    };
    FILE *exclude_file;
    char line_name[ FILENAME_SIZE ];
    char full_name[ FILENAME_SIZE ];

    while ( pAlways[ nSortedExcludeEntries ] )
        ++nSortedExcludeEntries;

    if ( !options.exclude )
        exclude_file = NULL;
    else
    {
        IsDir( options.exclude, 1, 1, NULL );
        exclude_file = fopen( options.exclude, "r" );
        if ( !exclude_file )
            FatalError( "Can't open exclude file: %s\n", options.exclude );
        while ( !feof( exclude_file ) && !ferror( exclude_file ) ) 
        {
            if ( !fgets( line_name, FILENAME_SIZE, exclude_file ) ||
                    line_name[0] == '\n' )
                continue;
            /* this is more hypothetical ... */
            if ( nSortedExcludeEntries == LONG_MAX )
                FatalError( "Too many entries in exclude file: %ld\n",
                        nSortedExcludeEntries );
            ++nSortedExcludeEntries;
        }
    }
    if ( nSortedExcludeEntries )
    {
        long j = 0;
        char *result;
        pSortedExcludeArray = 
            (char**) malloc( nSortedExcludeEntries * sizeof(char*) );
        if ( !pSortedExcludeArray )
            FatalError( "Can't allocate sorted array of exclude names.\n" );
        while ( pAlways[j] )
        {
            char *pDup = strdup( pAlways[j] );
            if ( !pDup )
                FatalError( "Can't allocate entry in sorted exclude array.\n" );
            pSortedExcludeArray[j] = PervertPath( pDup );
            ++j;
        }
        if ( exclude_file )
        {
            fseek( exclude_file, 0, SEEK_SET );
            for ( ; j < nSortedExcludeEntries && !feof( exclude_file )
                    && !ferror( exclude_file ); ++j ) 
            {
                int nLen;
                if ( !fgets( line_name, FILENAME_SIZE, exclude_file ) ||
                        line_name[0] == '\n' )
                {
                    --j;
                    continue;
                }
                nLen = strlen( line_name );
                line_name[ --nLen ] = '\0';     /* remove trailing linefeed */
                PervertPath( line_name );
                while ( nLen > 1 && line_name[nLen-1] == SEPARATOR[0] )
                    line_name[ --nLen ] = '\0'; /* remove trailing separator */
                pSortedExcludeArray[j] = strdup( line_name );
                if ( !pSortedExcludeArray[j] )
                    FatalError( "Can't allocate entry in sorted exclude array.\n" );
            }
            while ( (result = fgets( line_name, FILENAME_SIZE, exclude_file ))
                    && (line_name[0] == '\n') )
                ;   /* skip trailing empty lines */
            if ( j < nSortedExcludeEntries || result || !feof( exclude_file )
                    || ferror( exclude_file ) )
                FatalError( "Error in reading exclude file.\n" );
        }
        qsort( pSortedExcludeArray, nSortedExcludeEntries, sizeof(char*),
            &ExcludeSortFunction );
    }
    if ( exclude_file )
        fclose( exclude_file );
}


/* erAck:
 * Search a sorted array of exclude entries for a filename
 */
const char * SearchExcludeArray( const char* file_name )
{
    long hi, lo;
    hi = nSortedExcludeEntries - 1;
    lo = 0;
    while ( lo <= hi )
    {
        /* safer against overflows than ((lo+hi)/2) */
        long mid = lo + ((hi - lo) / 2);
        int n = FILENAMECMP( pSortedExcludeArray[mid], file_name );
        if ( n < 0 )
            lo = mid + 1;
        else if ( n > 0 )
            hi = mid - 1;
        else
            return( pSortedExcludeArray[mid] );
    }
    return( NULL );
}


void DeleteExcludeArray()
{
    if ( pSortedExcludeArray )
    {
        long j;
        for ( j=0; j < nSortedExcludeEntries; ++j )
        {
            free( pSortedExcludeArray[j] );
        }
        free( pSortedExcludeArray );
        pSortedExcludeArray = NULL;
        nSortedExcludeEntries = 0;
    }
}


void BuildOneArgument( FILE* crc_file, char* input_dir_name )
{
    char path[ FILENAME_SIZE ];
    FILE *list_file;
    int nLen;

    /* erAck: input_dir_name may be a @file containing a list of dirs */
    strcpy( path, input_dir_name );
    nLen = strlen( path );
    if ( nLen > 1 && path[0] == '@' )
    {
        /* must not stat if it doesn't exist */
        IsDir( &path[1], 1, 0, NULL );
        list_file = fopen( &path[1], "r" );
        if ( !list_file )
            printf( "Error: Can't open directory list file: %s\n", &path[1] );
            /* continue and try to take it as a real directory name */
    }
    else
        list_file = NULL;

    do
    {
        int restrict = options.restrict;
        if ( list_file )
        {
            if ( !fgets( path, FILENAME_SIZE, list_file ) )
            {
                if ( feof( list_file ) )
                    break;
                FatalError( "Could not read from directory list file: %s\n",
                        input_dir_name );
            }
            nLen = strlen( path );
            if ( nLen && path[nLen-1] == '\n' )
                path[--nLen] = '\0';
            if ( nLen > 4 && strcmp( &path[nLen-4], "  -r" ) == 0 )
            {
                options.restrict = 1;
                nLen -= 4;
                path[nLen] = '\0';
            }
            else if ( nLen > 5 && strcmp( &path[nLen-5], "  -r-" ) == 0 )
            {
                options.restrict = 0;
                nLen -= 5;
                path[nLen] = '\0';
            }
        }
        if ( nLen )
        {
            if ( path[ nLen - 1 ] == SEPARATOR[ 0 ] ||
                    path[ nLen - 1 ] == PERVERT_SEPARATOR[ 0 ] )
                path[ --nLen ] = '\0';
            PrependCwdPath( path );
            ProcessAllFiles( path, crc_file, 0 );
        }
        options.restrict = restrict;
    } while ( list_file && !feof( list_file ) );

    if ( list_file )
        fclose( list_file );
}


/*
 * Building the CRC file is a little harder than just checking
 * the file values.  This is because the build operation has to
 * parse through a directory tree, checking every file.  This
 * routine defers the hard part of that to a routine called
 * ProcessAllFiles(), which takes care of scanning through the
 * directory.  That means all we have to do here is open the output
 * CRC file, and then start the processing.  This routine also makes
 * sure that the directory name passed on the command line is
 * stripped of any trailing '/' or '\' character, since people tend
 * to include those when specifying directory names.
 */

void BuildCRCFile( char *crc_file_name, char **dir_names, int dir_count )
{
    FILE *crc_file;
    int nDir;

    /* must not stat if it doesn't exist */
    IsDir( crc_file_name, 1, 0, NULL );
    if ( options.pervert )
        crc_file = fopen( crc_file_name, (options.append ? "r+b" : "wb") );
    else
        crc_file = fopen( crc_file_name, (options.append ? "r+" : "w") );
    if ( crc_file == NULL )
        FatalError( "Can't open crc log file: %s\n", crc_file_name );
    if ( options.append )
        CreateCRCFileArray( crc_file );
    CreateExcludeArray();

    for ( nDir=0; nDir < dir_count; ++nDir )
        BuildOneArgument( crc_file, dir_names[nDir] );

    fclose( crc_file );
    DeleteCRCFileArray();
    DeleteExcludeArray();
}


/* erAck:
 * Check if a file name is a directory or not. Abort if file needed.
 */
int IsDir( const char* name, int file_needed, int must_stat, mode_t *pnStatMode )
{
    int ret, bad;
    struct stat buf;
    if ( pnStatMode )
        *pnStatMode = 0;
    bad = (stat( name, &buf ) == -1);
    if ( !bad )
    {
        ret = S_ISDIR(buf.st_mode);
        if ( pnStatMode )
            *pnStatMode = buf.st_mode;
    }
    else
    {
        int lbad = 0;
        ret = 0;
#if USE_LSTAT
        lbad = (lstat( name, &buf ) == -1);
        if ( !lbad && S_ISLNK(buf.st_mode) )
        {
            char target[ FILENAME_SIZE ];
            int len;
            if ( pnStatMode )
                *pnStatMode = buf.st_mode;
            OutInfo( "Warning: Dangling link? %s ", name );
            len = readlink( name, target, FILENAME_SIZE-1 );
            if ( len < 0 )
                OutInfo( "(can't resolve, %s)\n", strerror( errno ) );
            else
            {
                target[ len ] = '\0';
                OutInfo( "-> %s\n", target );
            }
        }
#endif // USE_LSTAT
        if ( must_stat && (file_needed || lbad) )
            FatalError( "Error reading stat from file %s: %s\n", name,
                    strerror( errno ) );
    }
    if ( ret && file_needed )
        FatalError( "Needed file, not directory: %s\n", name );
    return( ret );
}


/* erAck:
 * Check if a file exists under the given name.
 */
int Exists( const char* name )
{
    int ret;
    struct stat buf;
#if USE_POSIX_DIR
    ret = ( access( name, F_OK ) == 0 );
#else
    ret = ( stat( name, &buf ) == 0 );
#endif
#if USE_LSTAT
    if ( !ret )
        ret = (lstat( name, &buf ) == 0);
#endif // USE_LSTAT
    return ret;
}


/* erAck:
 * Find a name in the excluded list.
 */
const char * SearchExcludes( const char* pFullPathName,
        const char* pGivenName, const char* pFileName )
{
    const char * pEx;
    ((pEx = SearchExcludeArray( pFullPathName )) ||
     (pEx = SearchExcludeArray( pGivenName )) ||
     (pEx = SearchExcludeArray( pFileName ))
     );
    return pEx;
}


/*
 * This routine is responsible for actually performing the
 * calculation of the 32 bit CRC for the entire file.  We
 * precondition the CRC value with all 1's, then invert every bit
 * after the entire file has been done.  This gives us a CRC value
 * that corresponds with the values calculated by PKZIP and ARJ.
 * The actual calculation consists of reading in blocks from the
 * file, then updating the CRC with the value for that block.  The
 * CRC work is done by another the CalculateBufferCRC routine.
 */

unsigned long CalculateFileCRC( FILE* file )
{
    unsigned long crc;
    int count;
    unsigned char buffer[ 512 ];
    int i;

    crc = 0xFFFFFFFFL;
    i = 0;
    for ( ; ; ) {
        count = fread( buffer, 1, 512, file );
        if ( !options.quiet && ( i++ % 32 ) == 0 )
            putc( '.', stdout );
        if ( count == 0 )
            break;
        crc = CalculateBufferCRC( count, crc, buffer );
    }
    if ( !options.quiet )
        putc( ' ', stdout );
    return( crc ^= 0xFFFFFFFFL );
}


/* erAck:
 * Build the CRC for one file.
 * WriteCrcLog() may modify the name buffer passed!
 */
int ProcessOneFile( FILE* crc_file, char* name )
{
    int ret = 0;
    FILE* file = fopen( name, "rb" );
    if ( file )
    {
        unsigned long crc;
        OutProgress( "Scanning %s ", name );
        crc = CalculateFileCRC( file );
        if ( options.quiet < 1 )
            putc( '\n', stdout );
        WriteCrcLog( crc_file, crc, name );
        fclose( file );
        ret = 1;
    }
    else
    {
        struct stat buf;
        int fileerrno = errno;
#if USE_LSTAT
        int lbad = (lstat( name, &buf ) == -1);
        if ( !lbad && S_ISLNK(buf.st_mode) )
        {
            char target[ FILENAME_SIZE ];
            int len;
            OutInfo( "Scanning link only: %s ", name );
            len = readlink( name, target, FILENAME_SIZE-1 );
            if ( len < 0 )
            {
                fileerrno = errno;
                OutInfo( "(can't resolve, %s)\n", strerror( errno ) );
            }
            else
            {
                unsigned long crc = 0xFFFFFFFFL;
                target[ len ] = '\0';
                OutInfo( "-> %s\n", target );
                crc = CalculateBufferCRC( len, crc, target );
                /* NOT xor'ing out with 0xFFFFFFFFL to distinguish from
                 * identical file content */
                WriteCrcLog( crc_file, crc, name );
                ret = 1;
            }
        }
#endif // USE_LSTAT
        if ( !ret )
            FatalError( "\001Could not open %s: %s\n", name, strerror( errno ));
            /* \001: ignorable error */
    }
    return( ret );
}


/*
 * This is the routine that is responsible for calculating all of
 * the CRC values for the files in a given directory.  The CRC
 * values and the file names are written out to the crc_file. This
 * routine is somewhat ugly and hard to read because of all the
 * conditional code.  When searching through directories under
 * MS-DOS and UNIX, the flow of control is identical, but all of the
 * function calls, structure names, and elements are different.
 *
 * This routine sits in a loop for each directory, opening each file
 * and processing it.  Before a file is opened, a check is made to
 * see if the file is actually a directory.  If it turns out that
 * the file is a directory, a new path name is constructed, and this
 * routine calls itself recursively so that all the files in the
 * subdirectory are also processed.
 */

void ProcessAllFiles( char* path, FILE* crc_file, int bIsInRecursion )
{
#if USE_POSIX_DIR
    DIR *dirp;
#ifdef M_XENIX
    struct direct *entry;
#else
    struct dirent *entry;
#endif /* M_XENIX */
#define NAME entry->d_name
#else
    FILE_INFO fileinfo;
    int done;
#define NAME FILE_NAME( fileinfo )
#endif
    char fullname[ FILENAME_SIZE ];
    char full_path[ FILENAME_SIZE ];
    char *pPath;
    int  bOneFile;
    mode_t  nStatMode;

    PervertPath( path );
    GetFullPathName( full_path, path );
    pPath = (options.fullpath ? full_path : path);

    /* When is a file a file? Nothing is not a directory too, and don't try to
     * read from sockets or pipes. */
    bOneFile = !IsDir( pPath, 0, 0, &nStatMode ) &&
        (S_ISREG(nStatMode) || S_ISBLK(nStatMode) || S_ISCHR(nStatMode)
#ifdef S_ISLNK
         || S_ISLNK(nStatMode)
#endif
         );
    if ( !bOneFile && !S_ISDIR(nStatMode) )
    {
        OutInfo( "Warning: Don't know how to handle file type of %s, st_mode: 0x%lx\n",
                pPath, (unsigned long)nStatMode );
        return;
    }
    /* Recursion is always entered through a directory, bail out later if it
     * wasn't. */
    bOneFile = bOneFile && !bIsInRecursion;

    /* TODO: could extract the file name and use SearchExcludes(); right now
     * this may override a name-only exclude if a path was specified.
     * This part gets called unchecked for user specified path names only,
     * recursive calls already had the SearchExcludes() check, so this check
     * here can be avoided during recursion.  */

    if ( !bIsInRecursion &&
            (SearchExcludeArray( path ) || SearchExcludeArray( full_path )) )
    {
        OutInfo( "Excluding %s %s == %s\n", (bOneFile ? "file" : "directory"),
                path, full_path );
        return;
    }

    if ( bOneFile )
    {
        if ( !options.append || !SearchCRCFileArray( pPath, NULL ) )
            if ( !ProcessOneFile( crc_file, pPath ) )
                FatalError( "\001Error processing file %s\n", pPath );
                /* \001: ignorable error */
        return;
    }

    OutInfo( "Searching %s\n", pPath );
    strcat( pPath, SEPARATOR );
#if USE_POSIX_DIR
    dirp = opendir( pPath );
    if ( dirp == NULL )
    {
        /* \001: ignorable error */
        FatalError( "\001Error opening directory %s\n", pPath );
        if ( options.ignore )
            return;
    }
    entry = readdir( dirp );
    while ( entry != 0 )
    {
#else
    strcpy( fullname, pPath );
    strcat( fullname, "*.*" );
    done = FIND_FIRST( fullname, &fileinfo );
    while ( done == 0 )
    {
#endif
        if ( strcmp( NAME, "." ) && strcmp( NAME, ".." ) )
        {
            int bIsDir, bIsFile;
            strcpy( fullname, pPath );
            strcat( fullname, NAME );
            /* must not stat if ignorable */
            bIsDir = IsDir( fullname, 0, !options.ignore, &nStatMode );
            bIsFile = !bIsDir &&
                (S_ISREG(nStatMode) || S_ISBLK(nStatMode) || S_ISCHR(nStatMode)
#ifdef S_ISLNK
                 || S_ISLNK(nStatMode)
#endif
                );
            if ( bIsDir )
            {
                if ( !options.restrict )
                {
                    char full_file[ FILENAME_SIZE ];
                    const char* pEx;
                    PervertPath( fullname );
                    GetFullPathName( full_file, fullname );
                    pEx = SearchExcludes( full_file, fullname, NAME );
                    if ( pEx )
                        OutInfo( "Excluding directory %s == %s\n", pEx, full_file );
                    else
                        ProcessAllFiles( fullname, crc_file, 1 );
                }
            }
            else if ( bIsFile && (!options.append || !SearchCRCFileArray(
                            fullname, NULL )) )
            {
                char full_file[ FILENAME_SIZE ];
                const char* pEx;
                PervertPath( fullname );
                GetFullPathName( full_file, fullname );
                pEx = SearchExcludes( full_file, fullname, NAME );
                if ( pEx )
                    OutInfo( "Excluding file %s == %s\n", pEx, full_file );
                else
                    ProcessOneFile( crc_file, fullname );
            }
            else if ( !bIsDir && !bIsFile )
                OutInfo( "Not handling type of %s, st_mode: 0x%lx\n",
                        fullname, (unsigned long)nStatMode );
        }
#if USE_POSIX_DIR
        entry = readdir( dirp );
    }
    /* erAck: bugfix of original source
     * don't forget this, or you might run out of directory handles */
    closedir( dirp );   
#else
        done = FIND_NEXT( &fileinfo );
    }
#endif
}

/*
 * This routine calculates the CRC for a block of data using the
 * table lookup method. It accepts an original value for the crc,
 * and returns the updated value.
 */

unsigned long CalculateBufferCRC( unsigned int count, unsigned long crc,
        void* buffer )
{
    unsigned char *p;
    unsigned long temp1;
    unsigned long temp2;

    p = (unsigned char*) buffer;
    while ( count-- != 0 ) {
        temp1 = ( crc >> 8 ) & 0x00FFFFFFL;
        temp2 = CRCTable[ ( (int) crc ^ *p++ ) & 0xff ];
        crc = temp1 ^ temp2;
    }
    return( crc );
}

/*
 * The fatal error handler just has to print out a formatted error
 * message and then exit.  The difficulty in this routine lies in
 * the different way that variable numbers of arguments are
 * processed under ANSI C and K&R C.  The body of the routine is
 * the same under both environments, but the declaration of the
 * function and its arguments differs quite a bit.
 */

#ifdef USE_VARARGS

void FatalError( va_alist )
    va_dcl
{
    char *fmt;
    va_list argptr;

    va_start( argptr );
    fmt = va_arg( argptr, char * );

#else

void FatalError( char *fmt, ... )
{
    va_list argptr;
    va_start( argptr, fmt );

#endif

    if ( options.ignore && fmt[0] == '\001' )
    {
        printf( "Ignored Error: " );
        vprintf( fmt+1, argptr );
        va_end( argptr );
    }
    else
    {
        printf( "Fatal Error: " );
        vprintf( fmt, argptr );
        va_end( argptr );
        exit( -1 );
    }
}


/*
 * erAck:
 * Same mechanism for variadic output function. This is the Blah level.
 * Yes, I know, copy&paste, a quick hack.
 */

#ifdef USE_VARARGS

void OutProgress( va_alist )
    va_dcl
{
    char *fmt;
    va_list argptr;

    va_start( argptr );
    fmt = va_arg( argptr, char * );

#else

void OutProgress( char *fmt, ... )
{
    va_list argptr;
    va_start( argptr, fmt );

#endif

    if ( options.quiet < 1 )
        vprintf( fmt, argptr );
    va_end( argptr );
}


/*
 * and once again for the Info level
 */

#ifdef USE_VARARGS

void OutInfo( va_alist )
    va_dcl
{
    char *fmt;
    va_list argptr;

    va_start( argptr );
    fmt = va_arg( argptr, char * );

#else

void OutInfo( char *fmt, ... )
{
    va_list argptr;
    va_start( argptr, fmt );

#endif

    if ( options.quiet < 2 )
        vprintf( fmt, argptr );
    va_end( argptr );
}

/************************** End of CRCMAN.C ****************************/
